from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable


class ToHexStrMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'to_hex_str'
        args: dict[str, Variable] = {
            'data': Variable(Type.bytes)
        }
        super().__init__(identifier, args, return_type=Type.str)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal import constants
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        max_size_to_convert = constants.SIZE_OF_INT256
        hex_base = 16

        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(0)
        # if len(arg) == 0
        is_empty = code_generator.convert_begin_if()
        #   result = arg  # do nothing on script
        # else
        code_generator.change_jump(is_empty, Opcode.JMPEQ)

        #   no_calls = len(arg) // max_size + 1
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(max_size_to_convert)
        code_generator.convert_operation(BinaryOp.IntDiv, is_internal=True)
        code_generator.insert_opcode(Opcode.INC)
        #   result = ''
        code_generator.convert_literal('')
        #   index = 0
        code_generator.convert_literal(0)

        #   while index < no_calls:
        loop_start = code_generator.convert_begin_while()
        code_generator.convert_literal(hex_base)

        code_generator.duplicate_stack_item(5)
        #       pos = index * max_size
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(max_size_to_convert)
        code_generator.convert_operation(BinaryOp.Mul)

        #       subsize = min(len(arg) - pos, max_size)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Sub)
        code_generator.convert_literal(max_size_to_convert)
        code_generator.convert_builtin_method_call(Builtin.Min, is_internal=True)

        code_generator.convert_get_substring(is_internal=True, fix_result_type=False)
        #       aux = itoa(arg[pos: pos + max_size], hex_base)
        code_generator.convert_cast(Type.int, is_internal=True)
        code_generator.convert_builtin_method_call(Interop.Itoa, is_internal=True)

        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(2)
        code_generator.convert_operation(BinaryOp.Mod, is_internal=True)
        #       if len(aux) % 2:
        is_odd = code_generator.convert_begin_if()

        #           aux = '0' + aux
        code_generator.convert_literal('0')
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        code_generator.convert_end_if(is_odd)

        code_generator.swap_reverse_stack_items(3, rotate=True)
        #       result += aux
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.swap_reverse_stack_items(2)

        #       index += 1
        code_generator.insert_opcode(Opcode.INC)

        loop_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Gt)

        code_generator.convert_end_while(loop_start, loop_condition, is_internal=True)

        # clear stack
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)

        code_generator.convert_cast(self.return_type)

        code_generator.convert_end_if(is_empty)

    @property
    def _args_on_stack(self) -> int:
        return 0

    @property
    def _body(self) -> str | None:
        return None



from boa3.internal.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.internal.model.variable import Variable


class UnclaimedGasMethod(NeoContractMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'unclaimed_gas'
        native_identifier = 'unclaimedGas'
        args: dict[str, Variable] = {
            'account': Variable(UInt160Type.build()),
            'end': Variable(Type.int)
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.int)



import ast

from boa3.internal.model.builtin.interop.nativecontract import StdLibMethod
from boa3.internal.model.variable import Variable


class StringSplitMethod(StdLibMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'string_split'
        syscall = 'stringSplit'
        args: dict[str, Variable] = {
            'string': Variable(Type.str),
            'separator': Variable(Type.str),
            'remove_empty_entries': Variable(Type.bool),
        }

        args_default = ast.parse("{0}".format(False)
                                 ).body[0].value

        super().__init__(identifier, syscall, args, defaults=[args_default], return_type=Type.list.build([Type.str]))


class StringSplitWithoutRemoveEmptyEntriesMethod(StdLibMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'string_split'
        syscall = 'stringSplit'
        args: dict[str, Variable] = {
            'string': Variable(Type.str),
            'separator': Variable(Type.str),
        }

        args_default = ast.parse("{0}".format(False)
                                 ).body[0].value

        super().__init__(identifier, syscall, args, defaults=[args_default], return_type=Type.list.build([Type.str]))



from typing import Any, Self

from boa3.internal.model.method import Method
from boa3.internal.model.property import Property
from boa3.internal.model.type.classes.classarraytype import ClassArrayType
from boa3.internal.model.variable import Variable


class ContractPermissionType(ClassArrayType):
    """
    A class used to represent Neo ContractPermission class
    """

    def __init__(self):
        super().__init__('ContractPermission')
        from boa3.internal.model.type.type import Type

        self._variables: dict[str, Variable] = {
            'contract': Variable(Type.str),
            'methods': Variable(Type.optional.build(Type.list.build_collection([Type.str])))
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> dict[str, Method]:
        return {}

    def constructor_method(self) -> Method | None:
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> Self:
        if value is None or cls._is_type_of(value):
            return _ContractPermission

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, ContractPermissionType)


_ContractPermission = ContractPermissionType()



from boa3.internal.model.builtin.interop.nativecontract import StdLibMethod
from boa3.internal.model.variable import Variable


class StrLenMethod(StdLibMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'str_len'
        syscall = 'strLen'
        args: dict[str, Variable] = {
            'string': Variable(Type.str),
        }

        super().__init__(identifier, syscall, args, return_type=Type.int)



from typing import Any

from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.method import Method
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.bytestype import BytesType
from boa3.internal.model.type.type import Type
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class OpcodeType(BytesType):
    """
    A class used to represent Neo interop Opcode type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'Opcode'

    @property
    def default_value(self) -> Any:
        return Opcode.NOP

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if value is None or cls._is_type_of(value):
            return _Opcode

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (Opcode, OpcodeType))

    @property
    def symbols(self) -> dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the module with its name
        """
        _symbols = super().symbols
        _symbols.update({name: Variable(self) for name in Opcode.__members__.keys()})

        return _symbols

    def get_value(self, symbol_id) -> Any:
        """
        Gets the literal value of a symbol

        :return: the value if this type has this symbol. None otherwise.
        """
        if symbol_id in self.symbols:
            return Opcode.__members__[symbol_id]

        return None

    def constructor_method(self) -> Method | None:
        if self._constructor is None:
            self._constructor: Method = OpcodeMethod(self)
        return self._constructor


_Opcode = OpcodeType()


class OpcodeMethod(IBuiltinMethod):

    def __init__(self, return_type: OpcodeType):
        identifier = '-Opcode__init__'
        args: dict[str, Variable] = {
            'x': Variable(Type.bytes)
        }
        super().__init__(identifier, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 1

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp
        enum_values = [value for value in Opcode]

        for index, value in enumerate(enum_values):
            code_generator.duplicate_stack_item(code_generator.stack_size)
            code_generator.convert_literal(value)
            code_generator.convert_operation(BinaryOp.Eq, is_internal=True)
            if index > 0:
                code_generator.convert_operation(BinaryOp.Or, is_internal=True)

        code_generator.convert_literal(f"Invalid {self.return_type.identifier} parameter value")
        code_generator.convert_assert(has_message=True)



from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class StrBytesMultiplication(BinaryOperation):
    """
    A class used to represent a string or bytes concatenation operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.str, Type.bytes]

    def __init__(self, left: IType = Type.str, right: IType = Type.int):
        self.operator: Operator = Operator.Mult
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return any(_type.is_type_of(left) for _type in self._valid_types) and Type.int.is_type_of(right)

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return left
        else:
            return Type.none

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp

        # concatString = ''
        code_generator.convert_literal('')
        # reorganize stack
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)

        # while argInt > 0:
        concat_start = code_generator.convert_begin_while()

        #   concatString += argString
        code_generator.swap_reverse_stack_items(3)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        #   argInt -= 1
        code_generator.swap_reverse_stack_items(3)
        code_generator.insert_opcode(Opcode.DEC)

        # while condition
        condition_start = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)
        code_generator.convert_end_while(concat_start, condition_start, is_internal=True)

        # clear stack
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

        code_generator.convert_cast(self.left_type, is_internal=True)



__all__ = [
    'StackMemento',
    'NeoStack'
]

from boa3.internal.compiler.codegenerator.engine.istack import IStack
from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.VMCode import VMCode


class NeoStack(IStack):
    def __init__(self):
        from boa3.internal.model.type.itype import IType
        super().__init__(stack_type=IType)

    def _default_constructor_args(self) -> tuple:
        return tuple()


class StackMemento:
    """
    This class is responsible for managing the simulation of the blockchain stack during the code generation
    """

    def __init__(self):
        self._stacks: list[tuple[VMCode, NeoStack]] = []
        self._current_stack: NeoStack = NeoStack()

    @property
    def stack_map(self) -> dict[int, NeoStack]:
        vm_code_mapping = VMCodeMapping.instance()
        return {vm_code_mapping.get_start_address(vmcode): stack for vmcode, stack in self._stacks}

    def get_state(self, code_address: int) -> NeoStack:
        stacks = self.stack_map
        index = None
        for address, stack in sorted(stacks.items()):
            if address >= code_address:
                break
            index = address

        if index is None or index not in stacks:
            return NeoStack()
        else:
            return stacks[index]

    @property
    def current_stack(self) -> NeoStack:
        return self._current_stack

    def restore_state(self, code_address):
        stacks = self.stack_map
        latest_stack = None
        for address, stack in reversed(sorted(stacks.items())):
            if address < code_address:
                latest_stack = stack
                break

            vm_code = VMCodeMapping.instance().get_code(address)
            if (vm_code, stack) in self._stacks:
                self._stacks.remove((vm_code, stack))

        if latest_stack is not None:
            self._current_stack = latest_stack

    def append(self, code: VMCode, value: IType):
        states = self.stack_map
        index = VMCodeMapping.instance().get_start_address(code)
        if index in states:
            states[index].append(value)

        else:
            if self._current_stack is not None:
                stack = self._current_stack.copy()
            else:
                stack = NeoStack()
            stack.append(value)

            self._stacks.append((code, stack))
            self._current_stack = stack

    def pop(self, code: VMCode, index: int = -1):
        states = self.stack_map
        stack_index = VMCodeMapping.instance().get_start_address(code)
        if stack_index in states:
            stack = states[stack_index]
        else:
            if self._current_stack is not None:
                stack = self._current_stack.copy()
            else:
                stack = NeoStack()

            self._stacks.append((code, stack))
            self._current_stack = stack

        if len(stack) > 0:
            return stack.pop(index)

    def reverse(self, code: VMCode, start: int = 0, end: int = None, *, rotate: bool = False):
        states = self.stack_map
        stack_index = VMCodeMapping.instance().get_start_address(code)
        if stack_index in states:
            stack = states[stack_index]
        else:
            if self._current_stack is not None:
                stack = self._current_stack.copy()
            else:
                stack = NeoStack()

            self._stacks.append((code, stack))
            self._current_stack = stack

        return stack.reverse(start, end, rotate=rotate)



__all__ = [
    'NeoToken',
]


from typing import Any

from boa3.sc.types import ECPoint, UInt160, NeoAccountState
from boa3.sc.utils.iterator import Iterator


class NeoToken:
    """
    A class used to represent the NEO native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/Neo>`__
    to learn more about the NEO class.
    """

    hash: UInt160

    @classmethod
    def symbol(cls) -> str:
        """
        Gets the symbol of NEO.

        >>> NeoToken.symbol()
        'NEO'

        :return: the NEO string.
        :rtype: str
        """
        pass

    @classmethod
    def decimals(cls) -> int:
        """
        Gets the amount of decimals used by NEO.

        >>> NeoToken.decimals()
        0

        :return: the number 0.
        :rtype: int
        """
        pass

    @classmethod
    def totalSupply(cls) -> int:
        """
        Gets the total token supply deployed in the system.

        >>> NeoToken.totalSupply()
        100000000

        :return: the total token supply deployed in the system.
        :rtype: int
        """
        pass

    @classmethod
    def balanceOf(cls, account: UInt160) -> int:
        """
        Get the current balance of an address.

        >>> NeoToken.balanceOf(UInt160(bytes(20)))
        0

        >>> NeoToken.balanceOf(UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'))
        100

        :param account: the account's address to retrieve the balance for
        :type account: boa3.sc.types.UInt160
        :return: the account's balance
        :rtype: int
        """
        pass

    @classmethod
    def transfer(cls, from_address: UInt160, to_address: UInt160, amount: int, data: Any = None) -> bool:
        """
        Transfers an amount of GAS from one account to another.

        If the method succeeds, it will fire the `Transfer` event and must return true, even if the amount is 0,
        or from and to are the same address.

        >>> NeoToken.transfer(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'),     # this script hash needs to have signed the transaction or block
        ...              UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'),
        ...              10, None)
        True

        >>> NeoToken.transfer(UInt160(bytes(20)),
        ...              UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'),
        ...              10, None)
        False

        >>> NeoToken.transfer(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'),
        ...              UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'),
        ...              -1, None)
        False


        :param from_address: the address to transfer from
        :type from_address: boa3.sc.types.UInt160
        :param to_address: the address to transfer to
        :type to_address: boa3.sc.types.UInt160
        :param amount: the amount of NEO to transfer
        :type amount: int
        :param data: whatever data is pertinent to the onNEP17Payment method
        :type data: Any

        :return: whether the transfer was successful
        :rtype: bool
        :raise Exception: raised if `from_address` or `to_address` length is not 20 or if `amount` is less than zero.
        """
        pass

    @classmethod
    def get_gas_per_block(cls) -> int:
        """
        Gets the amount of GAS generated in each block.

        >>> NeoToken.get_gas_per_block()
        500000000

        :return: the amount of GAS generated
        :rtype: int
        """
        pass

    @classmethod
    def unclaimed_gas(cls, account: UInt160, end: int) -> int:
        """
        Gets the amount of unclaimed GAS in the specified account.

        >>> NeoToken.unclaimed_gas(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'), 0)
        100000000

        >>> NeoToken.unclaimed_gas(UInt160(bytes(20), 0)
        100000000

        :param account: the account to check
        :type account: boa3.sc.types.UInt160
        :param end: the block index used when calculating GAS
        :type end: int
        """
        pass

    @classmethod
    def register_candidate(cls, pubkey: ECPoint) -> bool:
        """
        Registers as a candidate.

        >>> NeoToken.register_candidate(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'))
        False

        :param pubkey: The public key of the account to be registered
        :type pubkey: boa3.sc.types.ECPoint
        :return: whether the registration was a success or not
        :rtype: bool
        """
        pass

    @classmethod
    def unregister_candidate(cls, pubkey: ECPoint) -> bool:
        """
        Unregisters as a candidate.

        >>> NeoToken.unregister_candidate(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'))
        False

        :param pubkey: The public key of the account to be unregistered
        :type pubkey: boa3.sc.types.ECPoint
        :return: whether the unregistration was a success or not
        :rtype: bool
        """
        pass

    @classmethod
    def vote(cls, account: UInt160, vote_to: ECPoint) -> bool:
        """
        Votes for a candidate.

        >>> NeoToken.vote(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'), ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'))
        False

        :param account: the account that is voting
        :type account: boa3.sc.types.UInt160
        :param vote_to: the public key of the one being voted
        :type vote_to: boa3.sc.types.ECPoint
        """
        pass

    @classmethod
    def get_all_candidates(cls) -> Iterator:
        """
        Gets the registered candidates iterator.

        >>> NeoToken.get_all_candidates()
        []

        :return: all registered candidates
        :rtype: Iterator
        """
        pass

    @classmethod
    def un_vote(cls, account: UInt160) -> bool:
        """
        Removes the vote of the candidate voted. It would be the same as calling vote(account, None).

        >>> NeoToken.un_vote(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'))
        False

        :param account: the account that is removing the vote
        :type account: boa3.sc.types.UInt160
        """
        pass

    @classmethod
    def get_candidates(cls) -> list[tuple[ECPoint, int]]:
        """
        Gets the list of all registered candidates.

        >>> NeoToken.get_candidates()
        []

        :return: all registered candidates
        :rtype: list[tuple[boa3.sc.types.ECPoint, int]]
        """
        pass

    @classmethod
    def get_candidate_vote(cls, pubkey: ECPoint) -> int:
        """
        Gets votes from specific candidate.

        >>> NeoToken.get_candidate_vote(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'))
        100

        >>> NeoToken.get_candidate_vote(ECPoint(bytes(32)))
        -1

        :return: Votes or -1 if it was not found.
        :rtype: int
        """
        pass

    @classmethod
    def get_committee(cls) -> list[ECPoint]:
        """
        Gets all committee members list.

        >>> NeoToken.get_committee()
        [ b'\\x02|\\x84\\xb0V\\xc2j{$XG\\x1em\\xcfgR\\xed\\xd9k\\x96\\x88}x34\\xe3Q\\xdd\\xfe\\x13\\xc4\\xbc\\xa2' ]

        :return: all committee members
        :rtype: list[boa3.sc.types.ECPoint]
        """
        pass

    @classmethod
    def get_committee_address(cls) -> UInt160:
        """
        Gets the address of the committee.

        >>> NeoToken.get_committee_address()
        UInt160(0x9273d3c792bce5eab4daac1c3ffdc1e83c4237f7)

        :return: the address of the committee
        :rtype: boa3.sc.types.UInt160
        """
        pass

    @classmethod
    def get_register_price(cls) -> int:
        """
        Gets the fees to be paid to register as a candidate.

        >>> NeoToken.get_register_price()
        100000000000

        :return: the amount of the fees
        :rtype: int
        """
        pass

    @classmethod
    def get_next_block_validators(cls) -> list[ECPoint]:
        """
        Gets validators list of the next block.

        >>> NeoToken.get_next_block_validators()
        [ b'\\x02|\\x84\\xb0V\\xc2j{$XG\\x1em\\xcfgR\\xed\\xd9k\\x96\\x88}x34\\xe3Q\\xdd\\xfe\\x13\\xc4\\xbc\\xa2' ]

        :return: the public keys of the validators
        :rtype: list[boa3.sc.types.ECPoint]
        """
        pass

    @classmethod
    def get_account_state(cls, account: UInt160) -> NeoAccountState:
        """
        Gets the latest votes of the specified account.

        >>> NeoToken.get_account_state(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'))
        {
            'balance': 100,
            'height': 2,
            'vote_to': None,
        }

        :param account: the specified account
        :type account: boa3.sc.types.UInt160
        :return: the state of the account
        :rtype: NeoAccountState
        """
        pass



from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.collection.sequence.tupletype import TupleType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class MaxMethod(IBuiltinMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        identifier = 'max'

        self._allowed_types = [Type.int, Type.str, Type.bytes]
        default_type = Type.int
        if not self._is_valid_type(arg_value):
            arg_value = default_type

        args: dict[str, Variable] = {
            'args1': Variable(arg_value),
            'args2': Variable(arg_value)
        }
        vararg = ('values', Variable(arg_value))
        super().__init__(identifier, args, return_type=arg_value, vararg=vararg)

    def _is_valid_type(self, arg_type: IType | None) -> bool:
        return (isinstance(arg_type, IType) and
                any(allowed_type.is_type_of(arg_type) for allowed_type in self._allowed_types))

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_values.type is Type.int:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_values.type._identifier)

    @property
    def _arg_values(self) -> Variable:
        return self._vararg[1]

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.MAX)

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.type.type import Type

        # if len(stack) == 2:
        code_generator.insert_opcode(Opcode.DEPTH, add_to_stack=[Type.int])
        code_generator.convert_literal(2)
        if_stack_size_equals_2 = code_generator.convert_begin_if()
        code_generator.change_jump(if_stack_size_equals_2, Opcode.JMPNE)

        #   aux_list = [arg1, arg2]
        code_generator.convert_literal(2)

        # else:
        else_stack_size_equals_3 = code_generator.convert_begin_else(if_stack_size_equals_2, is_internal=True)
        #   aux_list = [arg1, arg2, *values]
        code_generator.swap_reverse_stack_items(3)
        code_generator.insert_opcode(Opcode.UNPACK)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.convert_end_if(else_stack_size_equals_3, is_internal=True)
        code_generator.insert_opcode(Opcode.PACK, add_to_stack=[Type.list.build(self._arg_values.type)])

        # index = len(aux_list) - 1
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        # current_max = aux_list[index]
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)

        # while (index != 0):
        start_while = code_generator.convert_begin_while()

        #   index -= 1
        code_generator.swap_reverse_stack_items(2)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)

        #   current_max = max(current_max, aux_list[index])
        self._compare_values(code_generator)

        # while condition and end
        condition_address = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.insert_opcode(Opcode.SIGN)
        code_generator.convert_end_while(start_while, condition_address, is_internal=True)

        # return max
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

    def _compare_values(self, code_generator):
        """
        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        self.generate_internal_opcodes(code_generator)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) > 0:
            value = value[0]
        if isinstance(value, TupleType):
            value = value.value_type
        if type(value) == type(self._arg_values.type):
            return self

        from boa3.internal.model.builtin.method.maxbytestringmethod import MaxByteStringMethod
        from boa3.internal.model.builtin.method.maxintmethod import MaxIntMethod
        from boa3.internal.model.type.type import Type

        if Type.str.is_type_of(value) or Type.bytes.is_type_of(value):
            return MaxByteStringMethod(value)

        return MaxIntMethod(value)



from boa3.internal.model.builtin.method.printmethod import PrintMethod
from boa3.internal.model.type.classes.userclass import UserClass
from boa3.internal.model.type.itype import IType


class PrintClassMethod(PrintMethod):

    def __init__(self, arg_value: IType):
        if not isinstance(arg_value, UserClass):
            from boa3.internal.model.type.classes import userclass
            arg_value = userclass._EMPTY_CLASS

        super().__init__(arg_value)
        self._arg_type = arg_value

    def _generate_print_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        code_generator.convert_builtin_method_call(Builtin.StrClass.build(self._arg_type), is_internal=True)



from typing import Any

from boa3.internal.model.type.itype import IType


class EllipsisType(IType):
    """
    A class used to represent Python Ellipsis (...) annotation
    """

    def __init__(self):
        identifier = 'Ellipsis'
        super().__init__(identifier)

    @classmethod
    def build(cls, value: Any) -> IType:
        return ellipsisType

    @classmethod
    def _is_type_of(cls, value: Any):
        return value is Ellipsis or value is ellipsisType

    def union_type(self, other_type: IType) -> IType:
        return other_type

    def intersect_type(self, other_type: IType) -> IType:
        from boa3.internal.model.type.type import Type
        return Type.none


ellipsisType: IType = EllipsisType()



from boa3_test.test_drive.neoxp.command.neoexpresscommand.checkpoint.icheckpointcommand import ICheckpointCommand


class CheckpointRestoreCommand(ICheckpointCommand):
    def __init__(self, file_name: str | None = None,
                 force: bool = False,
                 neo_express_data_file: str = None):

        self.file_name = file_name
        self.force = force
        self.input = neo_express_data_file

        args = []
        if isinstance(file_name, str):
            args.append(file_name)

        super().__init__('restore', args)

    def _get_options(self) -> dict[str, str]:
        options = super()._get_options()

        if self.force:
            options['--force'] = ''

        return options



from abc import ABC
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.bytestype import BytesType
from boa3.internal.model.variable import Variable


class ToBoolMethod(IBuiltinMethod, ABC):
    def __init__(self, self_type: IType):
        identifier = 'to_bool'
        if isinstance(self_type, IdentifiedSymbol):
            identifier = '-{0}_{1}'.format(self_type.identifier, identifier)

        args: dict[str, Variable] = {'self': Variable(self_type)}
        from boa3.internal.model.type.type import Type
        super().__init__(identifier, args, return_type=Type.bool)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        return isinstance(params[0], IExpression) and isinstance(params[0].type, BytesType)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type
        code_generator.convert_cast(Type.bool, is_internal=True)

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None


class _ConvertToBoolMethod(ToBoolMethod):
    def __init__(self):
        super().__init__(None)

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, BytesType):
            return BytesToBoolMethod(value)
        # if it is not a valid type, show mismatched type with bytes
        return BytesToBoolMethod()


ToBool = _ConvertToBoolMethod()


class BytesToBoolMethod(ToBoolMethod):
    def __init__(self, self_type: IType = None):
        if not isinstance(self_type, BytesType):
            from boa3.internal.model.type.type import Type
            self_type = Type.bytes
        super().__init__(self_type)

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, BytesType):
            return BytesToBoolMethod(value)
        return super().build(value)



from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.type import IType, Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ListAddition(BinaryOperation):
    """
    A class used to represent a list addition operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.list]

    def __init__(self, left: IType = Type.list, right: IType = None):
        self.operator: Operator = Operator.Plus
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return left == right and any(_type.is_type_of(left) for _type in self._valid_types)

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return left
        else:
            return Type.none

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        # if isinstance(first, list):
        code_generator.duplicate_stack_item(2)
        code_generator.insert_type_check(StackItemType.Array)
        if_is_list = code_generator.convert_begin_if()

        #   aux = first
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_copy()
        code_generator.swap_reverse_stack_items(2)

        #   while there's values to add
        #   aux_index = len(second)
        values_type = self.right_type.value_type if hasattr(self.right_type, 'value_type') else Type.any
        code_generator.insert_opcode(Opcode.UNPACK, add_to_stack=[values_type, Type.int])

        #   while aux_index > 0:
        begin_while = code_generator.convert_begin_while()

        get_array_address = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.INC)    # push the array to the top of the stack
        code_generator.duplicate_stack_item(expected_stack_item=self.right_type)

        get_value_address = code_generator.bytecode_size
        code_generator.move_stack_item_to_top(3)

        #       aux.append(second[-aux_index])
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend,
                                                   [get_array_address, get_value_address],
                                                   is_internal=True)
        #       aux_index -= 1
        code_generator.insert_opcode(Opcode.DEC)

        #   while condition
        condition_start = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.convert_end_while(begin_while, condition_start, is_internal=True)

        #    clear stack
        code_generator.remove_stack_top_item()

        # else:
        else_is_list = code_generator.convert_begin_else(if_is_list, is_internal=True)
        #   simple concat
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.convert_end_if(else_is_list, is_internal=True)



__all__ = [
    'CryptoLib'
]

from boa3.sc.types import ECPoint, UInt160, NamedCurveHash, IBls12381


class CryptoLib:
    """
    A class used to represent the CryptoLib native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/CryptoLib>`__
    to learn more about the CryptoLib class.
    """

    hash: UInt160

    @classmethod
    def recover_secp256k1(cls, message_hash: bytes, signature: bytes) -> bytes | None:
        """
        Recovers the public key from a secp256k1 signature in a single byte array format.

        >>> CryptoLib.recover_secp256k1(
        ...     bytes.fromhex('5ae8317d34d1e595e3fa7247db80c0af4320cce1116de187f8f7e2e099c0d8d0'),
        ...     bytes.fromhex('45c0b7f8c09a9e1f1cea0c25785594427b6bf8f9f878a8af0b1abbb48e16d0920d8becd0c220f67c51217eecfd7184ef0732481c843857e6bc7fc095c4f6b78801')
        ... )
        b'\x03J\x07\x1e\x8an\x10\xaa\xda+\x8c\xf3\x9f\xa3\xb5\xfb4\x00\xb0N\x99\xea\x8a\xe6L\xee\xa1\xa9w\xdb\xea\xf5\xd5'

        >>> CryptoLib.recover_secp256k1(b'unit test', b'wrong signature')
        None

        :param message_hash: the hash of the message that was signed
        :type message_hash: bytes
        :param signature: the 65-byte signature in format: r[32] + s[32] + v[1]. 64-bytes for eip-2098, where v must be 27 or 28
        :type signature: bytes
        :return: the recovered public key in compressed format, or None if recovery fails.
        :rtype: bytes
        """
        pass

    @classmethod
    def murmur32(cls, data: bytes, seed: int) -> bytes:
        """
        Computes the hash value for the specified byte array using the murmur32 algorithm.

        >>> CryptoLib.murmur32(b'unit test', 0)
        b"\\x90D'G"

        :param data: the input to compute the hash code for
        :type data: bytes
        :param seed: the seed of the murmur32 hash function
        :type seed: int
        :return: the hash value
        :rtype: bytes
        """
        pass

    @classmethod
    def sha256(cls, data: bytes) -> bytes:
        """
        Encrypts a key using SHA-256.

        >>> CryptoLib.sha256(b'unit test')
        b'\\xdau1>J\\xc2W\\xf8LN\\xfb2\\x0f\\xbd\\x01\\x1cr@<\\xf5\\x93<\\x90\\xd2\\xe3\\xb8$\\xd6H\\x96\\xf8\\x9a'

        :param data: the data to be encrypted
        :type data: bytes
        :return: a byte value that represents the encrypted key
        :rtype: bytes
        """
        pass

    @classmethod
    def ripemd160(cls, data: bytes) -> bytes:
        """
        Encrypts a key using RIPEMD-160.

        >>> CryptoLib.ripemd160(b'unit test')
        b'H\\x8e\\xef\\xf4Zh\\x89:\\xe6\\xf1\\xdc\\x08\\xdd\\x8f\\x01\\rD\\n\\xbdH'

        :param data: the data to be encrypted
        :type data: Any
        :return: a byte value that represents the encrypted key
        :rtype: bytes
        """
        pass

    @classmethod
    def keccak256(cls, data: bytes) -> bytes:
        """
        Computes the hash value for the specified byte array using the keccak256 algorithm.

        >>> CryptoLib.keccak256(b'unit test')
        b'\\xe5\\x26\\x91\\x5a\\xff\\x6f\\x5e\\x35\\x9d\\x64\\xa3\\xce\\xf0\\x6e\\xf3\\xdb\\x9f\\x4a\\x89\\x0e\\x20\\xd1\\xa5\\x19\\x5e\\x3a\\x24\\x29\\x78\\x7e\\xec\\xb7'

        :param data: the input to compute the hash code for
        :type data: bytes
        :return: computed hash
        :rtype: bytes
        """
        pass

    @classmethod
    def verify_with_ecdsa(cls, message: bytes, pubkey: ECPoint, signature: bytes, curve: NamedCurveHash) -> bool:
        """
        Using the elliptic curve, it checks if the signature of the message was originally produced by the public key.

        >>> CryptoLib.verify_with_ecdsa(b'unit test', ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'),
        ...                             b'wrong_signature', NamedCurveHash.SECP256R1SHA256)
        False

        :param message: the encrypted message
        :type message: bytes
        :param pubkey: the public key that might have created the item
        :type pubkey: boa3.sc.types.ECPoint
        :param signature: the signature of the item
        :type signature: bytes
        :param curve: the curve that will be used by the ecdsa
        :type curve: boa3.sc.type.NamedCurveHash
        :return: a boolean value that represents whether the signature is valid
        :rtype: bool
        """
        pass

    @classmethod
    def bls12_381_add(cls, x: IBls12381, y: IBls12381) -> IBls12381:
        """
        Add operation of two bls12381 points.

        :param x: The first point
        :type x: IBls12381
        :param y: The second point
        :type y: IBls12381
        :return: the two points sum
        :rtype: IBls12381
        """
        pass

    @classmethod
    def bls12_381_deserialize(cls, data: bytes) -> IBls12381:
        """
        Deserialize a bls12381 point.

        :param data: The point as byte array
        :type data: bytes
        :return: the deserialized point
        :rtype: IBls12381
        """
        pass

    @classmethod
    def bls12_381_equal(cls, x: IBls12381, y: IBls12381) -> bool:
        """
        Determines whether the specified points are equal.

        :param x: The first point
        :type x: IBls12381
        :param y: The second point
        :type y: IBls12381
        :return: whether the specified points are equal or not
        :rtype: bool
        """
        pass

    @classmethod
    def bls12_381_mul(cls, x: IBls12381, mul: bytes, neg: bool) -> IBls12381:
        """
        Mul operation of gt point and multiplier.

        :param x: The point
        :type x: IBls12381
        :param mul: Multiplier, 32 bytes, little-endian
        :type mul: int
        :param neg: negative number
        :type neg: bool
        :return: the two points product
        :rtype: IBls12381
        """
        pass

    @classmethod
    def bls12_381_pairing(cls, g1: IBls12381, g2: IBls12381) -> IBls12381:
        """
        Pairing operation of g1 and g2.

        :param g1: The g1 point
        :type g1: IBls12381
        :param g2: The g2 point
        :type g2: IBls12381
        :return: the two points pairing
        :rtype: IBls12381
        """
        pass

    @classmethod
    def bls12_381_serialize(cls, g: IBls12381) -> bytes:
        """
        Serialize a bls12381 point.

        :param g: The point to be serialized.
        :type g: IBls12381
        :return: the serialized point
        :rtype: bytes
        """
        pass



from boa3.internal.model.builtin.interop.storage.getcheck.storagegetboolmethod import StorageTryGetBoolMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetbytesmethod import StorageTryGetBytesMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetdictmethod import StorageTryGetDictMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetecpointmethod import StorageTryGetECPointMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetintmethod import StorageTryGetIntMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetlistmethod import StorageTryGetListMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetobjectmethod import StorageTryGetObjectMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetstrmethod import StorageTryGetStrMethod
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetuint160method import StorageTryGetUInt160Method
from boa3.internal.model.builtin.interop.storage.getcheck.storagegetuint256method import StorageTryGetUInt256Method



from enum import IntEnum


class TransactionAttributeType(IntEnum):
    """
    Represents the TransactionAttributeType for running smart contracts.
    """

    HIGH_PRIORITY = 0x01
    """
    Indicates that the transaction is of high priority.
    
    :meta hide-value:
    """

    ORACLE_RESPONSE = 0x11
    """
    Indicates that the transaction is an oracle response

    :meta hide-value:
    """

    NOT_VALID_BEFORE = 0x20
    """
    Indicates that the transaction is not valid before height.

    :meta hide-value:
    """

    CONFLICTS = 0x21
    """
    Indicates that the transaction conflicts with hash.

    :meta hide-value:
    """



__all__ = [
    'literal_eval_node',
    'INVALID_NODE_RESULT'
]

import ast


class NodeEvaluationResult:
    pass


INVALID_NODE_RESULT = NodeEvaluationResult()


def literal_eval_node(node: ast.AST):
    try:
        return ast.literal_eval(node)
    except ValueError:
        return INVALID_NODE_RESULT



import abc
import ast
from collections.abc import Iterable, Sized
from typing import Any

from boa3.internal.model import set_internal_call
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class IStorageTryGetMethod(InteropMethod, abc.ABC):

    def __init__(self, identifier: str, value_type: IType):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.builtin.interop.storage.storagecontext.storagecontexttype import StorageContextType

        syscall = 'System.Storage.Get'
        context_type = StorageContextType.build()

        args: dict[str, Variable] = {'key': Variable(Type.bytes),
                                     'context': Variable(context_type)}

        from boa3.internal.model.builtin.interop.storage.storagegetcontextmethod import StorageGetContextMethod
        default_id = StorageGetContextMethod(context_type).identifier
        context_default = set_internal_call(ast.parse(f"{default_id}()"
                                                      ).body[0].value)

        return_type = Type.tuple.build((value_type, Type.bool))
        super().__init__(identifier, syscall, args, defaults=[context_default], return_type=return_type)

    @abc.abstractmethod
    def generate_default_value_opcodes(self, code_generator):
        pass

    @abc.abstractmethod
    def generate_deserialize_value_opcodes(self, code_generator):
        pass

    def generate_internal_opcodes(self, code_generator):
        super().generate_internal_opcodes(code_generator)

        from boa3.internal.model.operation.unaryop import UnaryOp

        # exists = result is not None
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(None)
        code_generator.convert_operation(UnaryOp.Not, is_internal=True)

        # if exists:
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_item(2)
        if_is_null = code_generator.convert_begin_if()

        #   result = deserialize(result)
        self.generate_deserialize_value_opcodes(code_generator)

        # else
        else_is_not_null = code_generator.convert_begin_else(if_is_null, is_internal=True)
        #   result = default_value
        code_generator.remove_stack_top_item()
        self.generate_default_value_opcodes(code_generator)

        if else_is_not_null < code_generator.last_code_start_address:
            if_is_null = else_is_not_null
        code_generator.convert_end_if(if_is_null, is_internal=True)

        code_generator.convert_new_array(2, self.return_type)

    @property
    def generation_order(self) -> list[int]:
        """
        Gets the indexes order that need to be used during code generation.
        If the order for generation is the same as inputted in code, returns reversed(range(0,len_args))

        :return: Index order for code generation
        """
        indexes = super().generation_order
        context_index = list(self.args).index('context')

        if indexes[-1] != context_index:
            # context must be the last generated argument
            indexes.remove(context_index)
            indexes.append(context_index)

        return indexes

    @property
    def key_arg(self) -> Variable:
        return self.args['key']

    def build(self, value: Any) -> IBuiltinMethod:
        exp: list[IExpression] = []
        if isinstance(value, Sized):
            if len(value) > 1 or not isinstance(value, Iterable):
                return self
            exp = [exp if isinstance(exp, IExpression) else Variable(exp)
                   for exp in value if isinstance(exp, (IExpression, IType))]

        elif isinstance(exp, (IExpression, IType)):
            exp = [value if isinstance(value, IExpression) else Variable(value)]
        else:
            return self

        if not self.validate_parameters(*exp):
            return self

        method = self
        key_type: IType = exp[0].type
        if method.key_arg.type.is_type_of(key_type):
            return self
        else:
            from boa3.internal.model.builtin.interop.storage.get.storagegetbytesmethod import StorageGetBytesMethod
            method = StorageGetBytesMethod()
            method.args['key'] = Variable(key_type)
        return method



from boa3.internal.model.builtin.interop.storage.get.storagegetboolmethod import StorageGetBoolMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetbytesmethod import StorageGetBytesMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetdictmethod import StorageGetDictMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetecpointmethod import StorageGetECPointMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetintmethod import StorageGetIntMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetlistmethod import StorageGetListMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetobjectmethod import StorageGetObjectMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetstrmethod import StorageGetStrMethod
from boa3.internal.model.builtin.interop.storage.get.storagegetuint160method import StorageGetUInt160Method
from boa3.internal.model.builtin.interop.storage.get.storagegetuint256method import StorageGetUInt256Method



from typing import Any

from boa3.internal.model.type.enum.intenumtype import IntEnumType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.type.ContractParameterType import ContractParameterType as ContractParameter


class ContractParameterTypeType(IntEnumType):
    """
    A class used to represent Neo interop ContractParameterType type
    """

    def __init__(self):
        super().__init__(ContractParameter)
        self._identifier = 'ContractParameterType'

    @property
    def default_value(self) -> Any:
        return ContractParameter.ALL

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if value is None or cls._is_type_of(value):
            return _ContractParameterType

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (ContractParameter, ContractParameterTypeType))


_ContractParameterType = ContractParameterTypeType()



from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand

__all__ = ['BatchCommand']


class BatchCommand(NeoExpressCommand):
    def __init__(self, file_name: str,
                 reset: bool = False,
                 check_point_file: str | None = None,
                 trace: bool = False,
                 neo_express_data_file: str = None):

        self.file_name = file_name
        self.reset = reset
        self.check_point_file = check_point_file
        self.trace = trace
        self.input = neo_express_data_file

        command_id = 'batch'
        super().__init__(command_id, [file_name])

    def _get_options(self) -> dict[str, str]:
        options = super()._get_options()

        if self.reset or isinstance(self.check_point_file, str):
            flag = '--reset'
            if isinstance(self.check_point_file, str):
                flag = f'{flag}:{self.check_point_file}'

            options[flag] = ''
        if self.trace:
            options['--trace'] = ''

        return options



from collections.abc import Iterable
from typing import Any

from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType


class TupleType(SequenceType):
    """
    A class used to represent Python tuple type
    """

    def __init__(self, values_type: list[IType] = None, any_length: bool = False):
        identifier = 'tuple'
        if values_type is None:
            values_type = []
            any_length = True

        self._tuple_types = values_type
        self._is_any_length = any_length

        values_type = self.filter_types(values_type)
        super().__init__(identifier, values_type)

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self.item_type == Type.any and self._is_any_length:
            return self._identifier

        if len(self._tuple_types) == 0:
            tuple_types = [self.item_type.identifier]
        else:
            tuple_types = [type_.identifier for type_ in self._tuple_types]

        if self._is_any_length:
            tuple_types.append('...')
        return f'{self._identifier}[{", ".join(tuple_types)}]'

    @property
    def default_value(self) -> Any:
        return tuple()

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.internal.model.type.type import Type
        return Type.int

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            values_types: list[IType] = cls.get_types(value)
            from boa3.internal.model.type.type import Type
            if len(values_types) == 2 and values_types[-1] is Type.ellipsis:
                has_ellipsis = True
                values_types.pop()
            else:
                has_ellipsis = False
                if Type.ellipsis in values_types:
                    # only tuple[<type>, ...] is accepted as tuple of any size typed as <type>
                    # all other cases where ... is used it has the same behavior as any
                    for index, value in enumerate(values_types):
                        if value is Type.ellipsis:
                            values_types[index] = Type.any

            return cls(values_types, any_length=has_ellipsis)

    def build_any_length(self, value: Any) -> IType:
        result: TupleType = self.build((value,))
        if len(result._tuple_types) == 1:
            result._is_any_length = True
        return result

    @classmethod
    def build_collection(cls, *value_type: IType | Iterable) -> IType:
        params = []
        for arg in value_type:
            if isinstance(arg, Iterable):
                argument = list(arg)
            else:
                argument = [arg]
            params.extend(argument)
        return cls.build(tuple(params))

    @classmethod
    def get_types(cls, value: Any) -> list[IType]:
        from boa3.internal.model.type.type import Type
        return [val if isinstance(val, IType) else Type.get_type(val) for val in value]

    def get_item_type(self, index: tuple):
        if len(index) > 0 and isinstance(index[0], int):
            target_index = index[0]
            if len(self._tuple_types) > target_index:
                return self._tuple_types[target_index]

        return super().get_item_type(index)

    def is_type_of(self, value: Any) -> bool:
        if self._is_type_of(value):
            min_size = len(self._tuple_types)
            if isinstance(value, TupleType):
                types_to_check = value._tuple_types
                any_length = value._is_any_length
            else:
                types_to_check = value
                any_length = False

            len_types_to_check = len(types_to_check)
            if self._is_any_length and len_types_to_check == 0 and not any_length:
                # tuples of any length are always type of empty tuple
                return True
            if len_types_to_check < min_size:
                return False
            if not self._is_any_length:
                if len_types_to_check > min_size:
                    return False
                elif len_types_to_check == min_size and any_length:
                    return False

            for index in range(min_size):
                if not self._tuple_types[index].is_type_of(types_to_check[index]):
                    return False
            if len_types_to_check > min_size:
                last_tuple_type = self._tuple_types[-1] if len(self._tuple_types) else self.value_type
                for index in range(min_size, len_types_to_check):
                    if not last_tuple_type.is_type_of(types_to_check[index]):
                        return False

            return True
        return False

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) is tuple or isinstance(value, TupleType)

    @property
    def can_reassign_values(self) -> bool:
        return False

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        return self.value_type == other.value_type

    def __hash__(self):
        return hash(self.identifier)



class _NotExecuted:
    def __repr__(self) -> str:
        return 'Invoke not executed yet'


class _TransactionExecuted:
    def __repr__(self) -> str:
        return 'Invoke was accepted and persisted. Use the transaction id to get the result'


class _Canceled:
    def __repr__(self) -> str:
        return 'Invoke was canceled before being executed'


NOT_EXECUTED = _NotExecuted()
CANCELED = _Canceled()
TRANSACTION_EXECUTED = _TransactionExecuted()



import abc
from typing import Any, Self

from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.wallet import utils


class Account(abc.ABC):
    def __init__(self, script_hash: UInt160,
                 account_version: int,
                 name: str = None,
                 label: str = None):

        if script_hash is not None:
            if isinstance(script_hash, bytes):
                script_hash = UInt160(script_hash)
            elif not isinstance(script_hash, UInt160):
                raise TypeError(f"Invalid data type {type(script_hash)}. Expecting UInt160 or bytes")
        self._script_hash: UInt160 = script_hash
        self._version: int = account_version

        if name is not None and not isinstance(name, str):
            raise TypeError(f"Invalid data type {type(name)}. Expecting str or None")
        self._name: str | None = name

        if label is not None and not isinstance(label, str):
            raise TypeError(f"Invalid data type {type(label)}. Expecting str or None")
        self._label: str | None = label

    @property
    def script_hash(self) -> UInt160:
        return self._script_hash

    @property
    def address(self) -> str | None:
        return utils.address_from_script_hash(self._script_hash.to_array(), self._version)

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def label(self) -> str | None:
        return self._label

    @classmethod
    @abc.abstractmethod
    def from_json(cls, json: dict[str, Any]) -> Self:
        pass

    def get_identifier(self) -> str:
        if isinstance(self._name, str):
            return self._name
        return self.address

    def __str__(self) -> str:
        prefix = f'{self._name}: ' if self._name is not None else ''
        return f'{prefix}{self.address}'

    def __repr__(self) -> str:
        return self.__str__()



from typing import Any, Self


class TransactionAttributeType:

    ORACLE_RESPONSE = 'OracleResponse'
    NOT_VALID_BEFORE = 'NotValidBefore'
    HIGH_PRIORITY_ATTRIBUTE = 'HighPriorityAttribute'


class TransactionAttribute:
    def __init__(self, type_: str):
        self._type = type_

    def to_json(self) -> dict[str, Any]:
        return {
            'type': self._type,
        }

    @classmethod
    def from_json(cls, json: dict[str, Any]) -> Self:
        if json['type'] == TransactionAttributeType.ORACLE_RESPONSE:
            from boa3_test.test_drive.model.network.payloads.oracleresponse import OracleResponse
            tx_attr = OracleResponse.from_json(json)
        else:
            tx_attr = TransactionAttribute(json['type'])

        return tx_attr



from typing import Any, Self

from boa3.internal.model.type.enum.intenumtype import IntEnumType
from boa3.internal.neo3.vm import VMState


class VMStateType(IntEnumType):
    """
    A class used to represent Neo interop VMState type
    """

    def __init__(self):
        super().__init__(VMState)
        self._identifier = 'VMState'

    @property
    def default_value(self) -> Any:
        return VMState.NONE

    @classmethod
    def build(cls, value: Any = None) -> Self:
        if value is None or cls._is_type_of(value):
            return _VMState

    @classmethod
    def _is_type_of(cls, value: Any = None):
        return isinstance(value, (VMState, VMStateType))

_VMState = VMStateType()
